/*  "hash cash" mint and remailer plug in for purposes of
 *  remailer postage (or any other use you can think of)
 * 
 *  See additional comments on development in hashcash.c
 *
 *  version 0.7ul: 02 June 1997   ant@notatla.demon.co.uk
 *               This is a library for easier merging with other code.
 *               Made less verbose, trimmed some code.
 *               Sender now chooses validity period, (start & duration)
 *               and remailer checks both.
 *
 *  version 0.8ul: 19 June 1997   ant@notatla.demon.co.uk
 *               File locking included on the database.
 *               Plus bugfixes: setting default name in right place
 *               and not calling sprintf(NULL,stuff).
 */

#include "hashcash.h"

void random_start (word32 *, word32 *, word32 *, word32 *);
void usage (void);
word32 find_collision (char day[5], char *, int, char *, word32, char *, int);
double usecs_for_n_hashes (word32, int);
/****  void hashcash_purge_database (char *,word32,char *);  ****/

/* a 120Mhz 486 does 28000 per second

   do a special case for tick = 1 second, say 5 seconds
   so 131071 tries
 */



hashcash_mint (char *base, int validity_period, int bits,
	       char *today, char *collision)
{
#define GROUP_SIZE ULONG_MAX
#define GROUP_DIGITS 8
#define GFORMAT "%08x"

#if 0
/* smaller base for debugging */
#define GROUP_SIZE 255
#define GROUP_DIGITS 2
#define GFORMAT "%02x"
#endif


  char counter[50];
  char *counter_ptr;
  word32 ran0, ran1, ran2, ran3;
  word32 i0, i1, i2;
  int found = 0;

  memset (counter, '\0', 37);
  random_start (&ran0, &ran1, &ran2, &ran3);
  timer_start ();
  counter_ptr = counter;

  for (i0 = 0; !found;)
    {
      sprintf (counter_ptr, GFORMAT GFORMAT, (i0 + ran0) & GROUP_SIZE,
	       ran1 & GROUP_SIZE);

      found = find_collision (today, base, bits, collision,
			      GROUP_SIZE, counter_ptr, validity_period);
      if (!found)
	{
	  i0 = (i0 + 1) & GROUP_SIZE;
	  if (i0 == 0)		/* ie if wrapped */
	    {
	      fprintf (stderr, "failed to find collision in 2^64 tries!\n");
	      exit (1);
	      /* shouldn't get here without trying  */
	      /* for a very long time */
	    }
	}
    }
  printf ("hashcash: %s\n", collision);
}
/*********************************************************/

#define DEFAULT_TRY ( ( tick_us() == 1000000 ) ? 131071 : 16383 )
word32
hashes_per_sec (int validity_period)
{
  word32 hashes_per_s = 0;
  double usecs;

  if (hashes_per_s == 0)
    {
      usecs = usecs_for_n_hashes (DEFAULT_TRY, validity_period);
      hashes_per_s = (word32) ((((double) DEFAULT_TRY) / usecs)
			       * 1000000.0);
    }
  return hashes_per_s;
}


double
usecs_for_n_hashes (word32 try, int validity_period)
{
  timer start;
  timer end;
  byte collision[20];
  char counter[50];
  double elapsed;

  timer_read (&start);

  sprintf (counter, GFORMAT GFORMAT, 0, 0);
  find_collision ("09856", "flame", 25, collision, try, counter, validity_period);

  timer_read (&end);

  elapsed = timer_interval (&start, &end);

  return elapsed;
}


int
count_collision_bits (char *base, char *check)
{
  char target[100];
  char try[100];
  SHA1_ctx ctx;
  byte target_digest[20];
  byte check_digest[20];
  int i;
  int last;
  int collision_bits;

  strncpy (target, check, 10);
  target[10] = '\0';
  strncat (target, base, sizeof (target));

  SHA1_init (&ctx);
  SHA1_update (&ctx, target, strlen (target));
  SHA1_final (&ctx);
  SHA1_digest (&ctx, target_digest);

  SHA1_init (&ctx);
  SHA1_update (&ctx, check, strlen (check));
  SHA1_final (&ctx);
  SHA1_digest (&ctx, check_digest);

  for (i = 0; i < 20 && check_digest[i] == target_digest[i]; i++);

  last = i;
  collision_bits = 8 * i;

#define bit( n, c ) (((c) >> (7 - (n))) & 1)

  for (i = 0; i < 8; i++)
    {
      if (bit (i, check_digest[last]) == bit (i, target_digest[last]))
	{
	  collision_bits++;
	}
      else
	{
	  break;
	}
    }
  return collision_bits;
}

void
random_start (word32 * w1, word32 * w2, word32 * w3, word32 * w4)
{
  /* kludge, might want to improve this, if collision collisions happen */

  timer t;
  word32 start;
  byte digest[20];

  SHA1_ctx ctx;

  SHA1_init (&ctx);

  timer_read (&t);

  SHA1_update (&ctx, &t, sizeof (timer));
  SHA1_final (&ctx);

  *w1 = ctx.H[0];
  *w2 = ctx.H[1];
  *w3 = ctx.H[2];
  *w4 = ctx.H[3];
}

word32
find_collision (char day[5], char *base, int bits,
		char *collision, word32 tries, char *counter,
		int validity_period)
{
  char *hex = "0123456789abcdef";
  char target[100];
  char try[100];
  char *changing_part_of_try;
  SHA1_ctx ctx;
  SHA1_ctx precomputed_ctx;
  word32 i;
  int j;
  word32 trial;
  word32 tries2;
  int len;
  int counter_len;
  int try_len;
  int try_strlen;
  byte target_digest[20];
  byte try_digest[20];
  int partial_byte = bits & 7;
  int check_bytes;
  int partial_byte_index;
  int partial_byte_mask;
  char last_char;
  int last_digit;

  memset (target, 0, 100);
  memcpy (target, day, 5);
  target[5] = 0;
  sprintf (target + 5, ":%03d:", validity_period);
  strncat (target, base, 95);

  counter_len = strlen (counter) - GROUP_DIGITS;
  sscanf (counter + counter_len, "%08x", &trial);
  trial -= trial % 16;		/* lop off last digit */

  if (partial_byte)
    {
      partial_byte_index = bits / 8;
      partial_byte_mask = ~((1 << (8 - (bits & 7))) - 1);
      check_bytes = partial_byte_index + 1;
    }
  else
    {
      check_bytes = bits / 8;
    }

  SHA1_init (&ctx);
  SHA1_update (&ctx, target, strlen (target));
  SHA1_final (&ctx);
  SHA1_digest (&ctx, target_digest);

  if (partial_byte)
    {
      target_digest[partial_byte_index] &= partial_byte_mask;
    }

  strncpy (try, target, 100);
  strncat (try, counter, counter_len);
  try_len = strlen (try);

/* length of try is fixed, GFORMAT is %08x, so move strlen outside loop */

  changing_part_of_try = try + try_len;
  sprintf (changing_part_of_try, GFORMAT, 0);
  try_strlen = strlen (try);

/* part of the ctx context can be precomputed as not all of the
   message is changing
 */

  tries2 = (int) ((double) tries / 16.0 + 0.5);
  for (i = 0; i < tries2; i++, trial = (trial + 16) & GROUP_SIZE)
    {
/* move precompute closer to the inner loop to precompute more */

      SHA1_init (&precomputed_ctx);

      sprintf (changing_part_of_try, GFORMAT, trial);

      SHA1_update (&precomputed_ctx, try, try_strlen - 1);

      for (j = 0; j < 16; j++)
	{
#if defined( DEBUG )
	  printf ("try: %s\n", try);
#endif
	  memcpy (&ctx, &precomputed_ctx, sizeof (SHA1_ctx));
	  last_char = hex[j];
	  SHA1_update (&ctx, &last_char, 1);
	  SHA1_final (&ctx);
	  SHA1_digest (&ctx, try_digest);
	  if (bits > 7)
	    {
	      if (try_digest[0] != target_digest[0])
		{
		  continue;
		}
	    }
	  if (partial_byte)
	    {
	      try_digest[partial_byte_index] &= partial_byte_mask;
	    }
	  if (memcmp (target_digest, try_digest, check_bytes) == 0)
	    {
	      changing_part_of_try[GROUP_DIGITS - 1] = hex[j];
	      strcpy (collision, try);
	      return i * 16 + j + 1;
	    }
	}
    }
  return 0;
}

#ifndef SPARSE
int
hashcash_create_if_nec (char *today, char *db_filename)
{
  FILE *db = fopen (db_filename, "r");
  if (NULL == db)
    {
      if (errno == ENOENT)
	{
	  fprintf (stderr, "DEV:%s:CREATION want some locking here!\n", db_filename);
	  fprintf (stderr, "DEV: Maybe I should study semaphores properly.\n");


	  db = fopen (db_filename, "w");
	  if (NULL == db)
	    {
	      perror (db_filename);
	      return 1;
	    }
	  fprintf (db, "%s\n", today);
	}
    }
  fclose (db);
  return 0;
}

int
hashcash_purge_if_nec (word32 today_num, char *today, char *db_filename)
{
  FILE *db;
  char *dbl;
  char day[] = "000000";

  if (hashcash_lock (db_filename))
    return 2;

  db = fopen (db_filename, "r");
  if (NULL == db)
    {
      perror (db_filename);
      return 3;
    }
  fgets (day, 6, db);
  day[5] = '\0';
  fclose (db);
  hashcash_unlock (db_filename);

  if (strncmp (day, today, 5))
/*** not purged already today ***/
    return hashcash_purge_database (today, today_num, db_filename);


  return 0;
}
/***************************************/

int
hashcash_add_to_database (word32 today_num, char *today, char *db_filename,
			  char *collision, int expiry)
{
/* dumb, DUMB, implementations of a database function - linear lookup */
  FILE *db;

  if (hashcash_create_if_nec (today, db_filename))
    return 2;
  if (hashcash_purge_if_nec (today_num, today, db_filename))
    return 2;


  if (hashcash_lock (db_filename))
    return 1;


  db = fopen (db_filename, "a");
  if (NULL == db)
    {
      perror (db_filename);
      return 1;
    }



  fseek (db, 0, SEEK_END);
  fprintf (db, "%d ", expiry);
  fputs (collision, db);
  fputc ('\n', db);
  fclose (db);
  hashcash_unlock (db_filename);
  return 0;
}

int
hashcash_in_database (word32 today_num, char *today,
		      char *db_filename, char *find)
{
  char entry[100];
  char *collision;
  char *expiry;
  int expiry_num;
  char day[6];
  word32 day_num;
  int len;
  FILE *db;

  if (hashcash_create_if_nec (today, db_filename))
    return 2;
  if (hashcash_purge_if_nec (today_num, today, db_filename))
    return 2;
  len = strlen (find);

  if (hashcash_lock (db_filename))
    return 2;
  db = fopen (db_filename, "r");
  if (NULL == db)
    {
      perror (db_filename);
      hashcash_unlock (db_filename);
      return 3;
    }

  fgets (entry, 100, db);	/* junk date line */
  while (!feof (db))
    {
      fgets (entry, 100, db);
      if (!feof (db))
	{
	  expiry = strtok (entry, " ");
	  collision = strtok (0, " ");
	  if (0 == strncmp (find, collision, len))
	    {
	      fclose (db);
	      hashcash_unlock (db_filename);
	      if (0 == expiry)
		return 0;
	      return 1;
	    }
	}
    }
  fclose (db);
  hashcash_unlock (db_filename);
  return 0;
}

int
hashcash_purge_database (char *today, word32 today_num, char *db_filename)
{
  char purged_filename[256];
  FILE *purged;
  int status;
  char entry[100];
  char day[6];
  word32 day_num;
  char *expiry;
  char *collision;
  word32 expiry_num;
  FILE *db = fopen (db_filename, "r");

  if (NULL == db)
    {
      perror (db_filename);
      return 1;
    }
  if (hashcash_lock (db_filename))
    return 1;

  strncpy (purged_filename, db_filename, 256);
  strcat (purged_filename, "1");

  purged = fopen (purged_filename, "w");
  if (NULL == purged)
    {
      perror (purged_filename);
      return 1;
    }

  fgets (entry, 100, db);	/* junk date line */
  while (!feof (db))
    {
      fgets (entry, 100, db);
      expiry = strtok (entry, " ");
      if (expiry == 0)
	{
	  continue;
	}
      collision = strtok (0, " ");
      if (collision == 0)
	{
	  continue;
	}
      expiry_num = atoi (expiry);
      strncpy (day, collision, 5);
      day[5] = '\0';
      day_num = atoi (day);
      if (expiry_num == 0 || expiry_num && day_num + expiry_num >= today_num)
	{
	  fputs (expiry, purged);
	  fputc (' ', purged);
	  fputs (collision, purged);
	}
    }
  fflush (purged);
  fclose (purged);
  fclose (db);
  unlink (db_filename);
  status = rename (purged_filename, db_filename);
  if (status < 0)
    {
      perror ("rename");
      return 1;
    }
  hashcash_unlock (db_filename);
  return 0;
}

void
usage (void)
{
  fprintf (stderr, "see http://www.dcs.ex.ac.uk/~aba/hashcash/ for details\n");
  fprintf (stderr, "\n\n\nThis version has been fiddled with by\n");
  fprintf (stderr, "Antonomasia <ant@notatla.demon.co.uk>\n");
  fprintf (stderr, "hashcash -t\n");
  fprintf (stderr, "hashcash -m  bits base [days] [now]\n");
  fprintf (stderr, "hashcash -s  bits base collision [maxdays] [-dname]\n");
  exit (1);
}
#endif
/*******************************************/


int 
hashcash_spend (word32 today_num, char *today, char *check,
		char *collision,
		int validity_period, int maxdays,
		int bits,
		int database_flag,
		char *db_filename)
{
  char duration[4];
  char day[6];
  int duration_num, collision_bits, in_db;

  word32 day_num;

  today_num = (time (0) / 3600) / 24;

  strncpy (day, collision, 5);
  day[5] = '\0';
  day_num = atoi (day);

  strncpy (duration, collision + 6, 3);
  duration[3] = '\0';
  duration_num = atoi (duration);


/** is the start still future ? **/
  if (day_num > today_num)
    return 1;

/** is the duration too long ? **/
  if ((maxdays > 0) && (duration_num > maxdays))
    return 1;

/** is the end past ? **/
  if ((day_num + duration_num) < today_num)
    return 1;


/** is there a long enough collision ? **/
  collision_bits = count_collision_bits (check, collision);
  printf ("required %d bits, got %d\n", bits, collision_bits);
  if (collision_bits < bits)
    return 1;

/** if NOT using the database we have finished **/
  if (!database_flag)
    return 0;

  if (NULL == db_filename)
    db_filename = strdup ("hashcash.db");
  if (strlen (db_filename) < 1)
    db_filename = strdup ("hashcash.db");

  in_db = hashcash_in_database (today_num, today, db_filename, collision);	/* if it's in there, it's a failure */
  if (1 == in_db)
    {
      printf ("double spent\n");
      return (1);
    }
  if (in_db)
    {
      printf ("Rejecting after trying to check database.\n");
      return (1);
    }

  return
    hashcash_add_to_database (today_num,
			      today, db_filename,
			      collision, validity_period);
}
/********************/
int 
hashcash_lock (char *db_filename)
{
  int i, try;
  char *dbl;

  dbl = (char *) malloc ((size_t) strlen (db_filename) + 6);
  if (NULL == dbl)
    {
      syslog (LOG_DAEMON, "hashcash library - Failed to lock database.\n");
      return 1;
    }

  sprintf (dbl, "%s.lock", db_filename);

  try = 1;
  for (i = 9; (i && try); i--)
    {
      try = link (db_filename, dbl);
      usleep ((unsigned long) 50);
    }
  if (try)
    {
      fprintf (stderr, "Aborting in hashcash_lock\n");
      abort ();
    }
  return try;

}
/**********************/
int 
hashcash_unlock (char *db_filename)
{
  char *dbl;
  if (NULL == db_filename)
    {
      fprintf (stderr, "Ignoring null filename in hashcash_unlock\n");
      return 0;
    }
  dbl = (char *) malloc ((size_t) strlen (db_filename) + 6);
  sprintf (dbl, "%s.lock", db_filename);
  unlink (dbl);
}
